/*
 * Vifir - An electronic flight bag wannabe utility for aircraft pilots
 * Copyright (C) 2009 David Ludlow
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * dave@adsllc.com, PO Box 13070, Toledo, OH 43623
 */

#define __STDC_LIMIT_MACROS
#include <stdint.h>
#include <cstdlib>
#include <cstring>
#include <cstdio>
#include <sstream>
#include <string>
#include <gtkmm/messagedialog.h>
#include <gtkmm/main.h>

#include <errno.h>
#ifdef WIN32
#include <process.h>
#endif

#include "ApproachPlate.h"
#include "Util.h"
#include "DataStore.h"

ApproachPlate::ApproachPlate()
{
	Cycle="";
	Effective=0;
	Expire=0;
	StateID="";
	StateName="";
	CityName="";
	Volume="";
	AirportID="";
	AirportMilitary=false;
	AirportIdent="";
	AirportICAOIdent="";
	Seq=0;
	Code=CODEINVALID;
	Name="";
	UserAction=ACTIONINVALID;;
	ImgName="";
	ImgType=PDF;
	CNFlag=false;
	CNSection=SECTIONINVALID;
	CNPage=0;
	BVSection=SECTIONINVALID;
	BVPage=0;
	ProcUID=0;
	TwoColored=false;
	Civil=CIVILINVALID;
	FAANFD15="";
	FAANFD18="";
	Copter=false;
}

ApproachPlate::~ApproachPlate()
{
}

void ApproachPlate::SetCycle(std::string Cycle) {
	this->Cycle=Cycle;
}

void ApproachPlate::SetEffective(time_t Effective) {
	this->Effective=Effective;
}

void ApproachPlate::SetExpire(time_t Expire) {
	this->Expire=Expire;
}

void ApproachPlate::SetStateID(std::string StateID) {
	this->StateID=StateID;
}

void ApproachPlate::SetStateName(std::string StateName) {
	this->StateName=StateName;
}

void ApproachPlate::SetCityName(std::string CityName) {
	this->CityName=CityName;
}

void ApproachPlate::SetVolume(std::string Volume) {
	this->Volume=Volume;
}

void ApproachPlate::SetAirportID(std::string AirportID) {
	this->AirportID=AirportID;
}

void ApproachPlate::SetAirportMilitary(std::string AirportMilitary) {
	if (AirportMilitary.length()==1) {
		if (AirportMilitary[0]=='M') {
			SetAirportMilitary(1);
		} else if (AirportMilitary[0]=='N') {
			SetAirportMilitary(0);
		}
	}
}

void ApproachPlate::SetAirportMilitary(bool AirportMilitary) {
	this->AirportMilitary=AirportMilitary;
}

void ApproachPlate::SetAirportIdent(std::string AirportIdent) {
	this->AirportIdent=AirportIdent;
}

void ApproachPlate::SetAirportICAOIdent(std::string AirportICAOIdent) {
	this->AirportICAOIdent=AirportICAOIdent;
}

void ApproachPlate::SetSeq(std::string Seq) {
	SetSeq(atoi(Seq.c_str()));
}

void ApproachPlate::SetSeq(int Seq) {
	this->Seq=Seq;
}

void ApproachPlate::SetCode(std::string Code) {
	if (Code=="MIN")
		this->Code=MIN;
	else if (Code=="STAR")
		this->Code=STAR;
	else if (Code=="IAP")
		this->Code=IAP;
	else if (Code=="APD")
		this->Code=APD;
	else if (Code=="DP")
		this->Code=DP;
	else if (Code=="DPO")
		this->Code=DPO;
	else {
		this->Code=CODEINVALID;
		fprintf(stderr,"Invalid chart code='%s'\n",Code.c_str());
	}
}

void ApproachPlate::SetCode(ChartCodeEnum Code) {
	this->Code=Code;
}

void ApproachPlate::SetName(std::string Name) {
	this->Name=Name;
}

void ApproachPlate::SetUserAction(std::string UserAction) {
	if (UserAction.length()==1)
		SetUserAction(UserAction[0]);
}

void ApproachPlate::SetUserAction(char UserAction) {
	if (UserAction=='A')
		this->UserAction=ADD;
	else if (UserAction=='C')
		this->UserAction=CHANGE;
	else if (UserAction=='D')
		this->UserAction=REMOVE;
	else if (UserAction==0)
		this->UserAction=UNCHANGED;
	else {
		this->UserAction=ACTIONINVALID;
	}
}

void ApproachPlate::SetUserAction(UserActionEnum UserAction) {
	this->UserAction=UserAction;
}

void ApproachPlate::SetImgName(std::string ImgName) {
	this->ImgName=ImgName;
}

void ApproachPlate::SetImgType(enum ImgTypeEnum ImgType) {
	this->ImgType=ImgType;
}

void ApproachPlate::SetCNFlag(std::string CNFlag) {
	if (CNFlag.length()==1) {
		if (CNFlag[0]=='Y')
			SetCNFlag(1);
		else if (CNFlag[0]=='N')
			SetCNFlag(0);
	}
}

void ApproachPlate::SetCNFlag(bool CNFlag) {
	this->CNFlag=CNFlag;
}

void ApproachPlate::SetCNSection(std::string CNSection) {
	if (CNSection.length()==1)
		SetCNSection(CNSection[0]);
}

void ApproachPlate::SetCNSection(char CNSection) {
	if (CNSection=='B')
		this->CNSection=TOMIN;
	else if (CNSection=='C')
		this->CNSection=ALTMIN;
	else if (CNSection=='D')
		this->CNSection=RADMIN;
	else if (CNSection=='E')
		this->CNSection=STARS;
	else {
		this->CNSection=SECTIONINVALID;
		fprintf(stderr,"Invalid Change Notice Section code='%c'\n",UserAction);
	}
}

void ApproachPlate::SetCNSection(SectionEnum CNSection) {
	this->CNSection=CNSection;
}

void ApproachPlate::SetCNPage(std::string CNPage) {
	SetCNPage(atoi(CNPage.c_str()));
}

void ApproachPlate::SetCNPage(int CNPage) {
	this->CNPage=CNPage;
}

void ApproachPlate::SetBVSection(std::string BVSection) {
	if (BVSection.length()==1)
		SetBVSection(BVSection[0]);
}

void ApproachPlate::SetBVSection(char BVSection) {
	if (BVSection=='C')
		this->BVSection=TOMIN;
	else if (BVSection=='E')
		this->BVSection=ALTMIN;
	else if (BVSection=='N')
		this->BVSection=RADMIN;
	else if (BVSection=='P')
		this->BVSection=STARS;
	else {
		this->BVSection=SECTIONINVALID;
		if (BVSection!=' ' && BVSection!='Z')
			fprintf(stderr,"Invalid Bound Volume Section code=%c\n",BVSection);
	}
}

void ApproachPlate::SetBVSection(SectionEnum BVSection) {
	this->BVSection=BVSection;
}

void ApproachPlate::SetBVPage(std::string BVPage) {
	SetBVPage(atoi(BVPage.c_str()));
}

void ApproachPlate::SetBVPage(int BVPage) {
	this->BVPage=BVPage;
}

void ApproachPlate::SetProcUID(std::string ProcUID) {
	SetProcUID(atoi(ProcUID.c_str()));
}

void ApproachPlate::SetProcUID(int ProcUID) {
	this->ProcUID=ProcUID;
}

void ApproachPlate::SetTwoColored(std::string TwoColored) {
	if (TwoColored.length()==1) {
		if (TwoColored[0]=='Y')
			SetTwoColored(1);
		else if (TwoColored[0]=='N')
			SetTwoColored(0);
	}
}

void ApproachPlate::SetTwoColored(bool TwoColored) {
	this->TwoColored=TwoColored;
}

void ApproachPlate::SetCivil(std::string Civil) {
	if (Civil.length()==1)
		SetCivil(Civil[0]);
}

void ApproachPlate::SetCivil(char Civil) {
	if (Civil=='C')
		this->Civil=CIVIL;
	else if (Civil=='D')
		this->Civil=DMAAC;
	else if (Civil=='N')
		this->Civil=NGA;
	else if (Civil=='H')
		this->Civil=NGAHIGH;
	else
	{
		this->Civil=CIVILINVALID;
	}
}

void ApproachPlate::SetCivil(CivilEnum Civil) {
	this->Civil=Civil;
}

void ApproachPlate::SetFAANFD15(std::string FAANFD15) {
	this->FAANFD15=FAANFD15;
}

void ApproachPlate::SetFAANFD18(std::string FAANFD18) {
	this->FAANFD18=FAANFD18;
}

void ApproachPlate::SetCopter(std::string Copter) {
	if (Copter.length()==1) {
		if (Copter[0]=='Y')
			SetCopter(1);
		else if (Copter[0]=='N')
			SetCopter(0);
	}
}

void ApproachPlate::SetCopter(bool Copter) {
	this->Copter=Copter;
}

std::string ApproachPlate::GetCycle() {
	return Cycle;
}

time_t ApproachPlate::GetEffective() {
	return Effective;
}

time_t ApproachPlate::GetExpire() {
	return Expire;
}

std::string ApproachPlate::GetStateID() {
	return StateID;
}

std::string ApproachPlate::GetStateName() {
	return StateName;
}

std::string ApproachPlate::GetCityName() {
	return CityName;
}

std::string ApproachPlate::GetVolume() {
	return Volume;
}

std::string ApproachPlate::GetAirportID() {
	return AirportID;
}

bool ApproachPlate::GetAirportMilitary() {
	return AirportMilitary;
}

std::string ApproachPlate::GetAirportIdent() {
	return AirportIdent;
}

std::string ApproachPlate::GetAirportICAOIdent() {
	return AirportICAOIdent;
}

int ApproachPlate::GetSeq() {
	return Seq;
}

ApproachPlate::ChartCodeEnum ApproachPlate::GetCode() {
	return Code;
}

std::string ApproachPlate::GetName() {
	return Name;
}

ApproachPlate::UserActionEnum ApproachPlate::GetUserAction() {
	return UserAction;
}

std::string ApproachPlate::GetImgName() {
	return ImgName;
}

enum ApproachPlate::ImgTypeEnum ApproachPlate::GetImgType() {
	return ImgType;
}

bool ApproachPlate::GetCNFlag() {
	return CNFlag;
}

ApproachPlate::SectionEnum ApproachPlate::GetCNSection() {
	return CNSection;
}

int ApproachPlate::GetCNPage() {
	return CNPage;
}

ApproachPlate::SectionEnum ApproachPlate::GetBVSection() {
	return BVSection;
}

int ApproachPlate::GetBVPage() {
	return BVPage;
}

int ApproachPlate::GetProcUID() {
	return ProcUID;
}

bool ApproachPlate::GetTwoColored() {
	return TwoColored;
}

ApproachPlate::CivilEnum ApproachPlate::GetCivil() {
	return Civil;
}

std::string ApproachPlate::GetFAANFD15() {
	return FAANFD15;
}

std::string ApproachPlate::GetFAANFD18() {
	return FAANFD18;
}

bool ApproachPlate::GetCopter() {
	return Copter;
}

void ApproachPlate::Copy(ApproachPlate *Other) {
	this->Cycle=Other->GetCycle();
	this->Effective=Other->GetEffective();
	this->StateID=Other->GetStateID();
	this->StateName=Other->GetStateName();
	this->CityName=Other->GetCityName();
	this->Volume=Other->GetVolume();
	this->AirportID=Other->GetAirportID();
	this->AirportMilitary=Other->GetAirportMilitary();
	this->AirportIdent=Other->GetAirportIdent();
	this->AirportICAOIdent=Other->GetAirportICAOIdent();
	this->Seq=Other->GetSeq();
	this->Code=Other->GetCode();
	this->Name=Other->GetName();
	this->UserAction=Other->GetUserAction();
	this->ImgName=Other->GetImgName();
	this->ImgType=Other->GetImgType();
	this->CNFlag=Other->GetCNFlag();
	this->CNSection=Other->GetCNSection();
	this->CNPage=Other->GetCNPage();
	this->BVSection=Other->GetBVSection();
	this->BVPage=Other->GetBVPage();
	this->ProcUID=Other->GetProcUID();
	this->TwoColored=Other->GetTwoColored();
	this->Civil=Other->GetCivil();
	this->FAANFD15=Other->FAANFD15;
	this->FAANFD18=Other->FAANFD18;
	this->Copter=Other->Copter;
}


std::string ApproachPlate::toUpper(std::string str) {
	for (unsigned int i=0;i<str.length();i++) {
		str[i] = toupper(str[i]);
	}
	return str;
}

bool ApproachPlate::Matchi(std::string str) {
	str=toUpper(str);

	if (
		(toUpper(AirportIdent).substr(0,str.length())==str) ||
		(toUpper(AirportICAOIdent).substr(0,str.length())==str) ||
		(toUpper(CityName).find(str)!=std::string::npos)
	)
	{
		return true;
	}
	else
	{
		return false;
	}
}

void ApproachPlate::View(DataStore *Data) {
	std::string Filename;

	Filename=(*(Data->GetActiveVolume()))->GetDirectory()+"/plateimages/"+ImgName;
	Util::ShellExpand(&Filename);

#ifndef WIN32
	pid_t forkresult;
	forkresult=fork();

	if (forkresult==-1) {
		std::ostringstream ErrMsg;
		ErrMsg<<"Unable to fork(): "<<strerror(errno)<< " ("<<errno<<")";
		Gtk::MessageDialog *message = new Gtk::MessageDialog(ErrMsg.str(),
			false,Gtk::MESSAGE_ERROR,Gtk::BUTTONS_CANCEL,true);
		message->run();
		delete message;

		return;
	}
	if (forkresult==0) {
		execlp("evince","evince",Filename.c_str(),(char *) NULL);
		fprintf(stderr,"Error executing '%s \"%s\"': %s (%u)\n","evince",Filename.c_str(),strerror(errno),errno);
		Gtk::Main::quit();
		return;
	}
#else
	spawnlp(P_NOWAIT,"acroread.exe","acroread.exe",Filename.c_str(),(char *) NULL);
#endif
}

bool ApproachPlate::MakeBinary(struct ApproachPlate_Binary_struct *BinaryData) {
	/* Sanity checking */
	if (
		StateID.size()>sizeof(BinaryData->StateID) ||
		StateName.size()>sizeof(BinaryData->StateName) ||
		CityName.size()>sizeof(BinaryData->CityName) ||
		Seq>UINT32_MAX ||
		Code>UINT8_MAX ||
		Name.size()>sizeof(BinaryData->Name) ||
		UserAction>UINT8_MAX ||
		ImgName.size()>sizeof(BinaryData->ImgName) ||
		ImgType>UINT8_MAX ||
		/* CNFlag - Nothing to check for booleans */
		CNSection>UINT8_MAX ||
		CNPage>UINT16_MAX ||
		BVSection>UINT8_MAX ||
		BVPage>UINT16_MAX ||
		ProcUID>UINT32_MAX ||
		/* TwoColored - Nothing to check for booleans */
		Civil>UINT8_MAX
		) {
		return false;
	}

	strncpy(BinaryData->StateID,StateID.c_str(),sizeof(BinaryData->StateID));
	strncpy(BinaryData->StateName,StateName.c_str(),sizeof(BinaryData->StateName));
	strncpy(BinaryData->CityName,CityName.c_str(),sizeof(BinaryData->CityName));
	BinaryData->Seq=Seq;
	BinaryData->Code=Code;
	strncpy(BinaryData->Name,Name.c_str(),sizeof(BinaryData->Name));
	BinaryData->UserAction=UserAction;
	strncpy(BinaryData->ImgName,ImgName.c_str(),sizeof(BinaryData->ImgName));
	BinaryData->ImgType=ImgType;
	BinaryData->CNFlag=CNFlag;
	BinaryData->CNSection=CNSection;
	BinaryData->CNPage=CNPage;
	BinaryData->BVSection=BVSection;
	BinaryData->BVPage=BVPage;
	BinaryData->ProcUID=ProcUID;
	BinaryData->TwoColored=TwoColored;
	BinaryData->Civil=Civil;

	return true;
}

void ApproachPlate::SetFromBinary(struct ApproachPlate_Binary_struct *BinaryData) {
	StateID.assign(BinaryData->StateID,sizeof(BinaryData->StateID));
	StateName.assign(BinaryData->StateName,sizeof(BinaryData->StateName));
	CityName.assign(BinaryData->CityName,sizeof(BinaryData->CityName));
	Seq=BinaryData->Seq;
	Code=(ApproachPlate::ChartCodeEnum)BinaryData->Code;
	Name.assign(BinaryData->Name,sizeof(BinaryData->Name));
	UserAction=(ApproachPlate::UserActionEnum)BinaryData->UserAction;
	ImgName.assign(BinaryData->ImgName,sizeof(BinaryData->ImgName));
	ImgType=(ApproachPlate::ImgTypeEnum)BinaryData->ImgType;
	CNFlag=BinaryData->CNFlag;
	CNSection=(ApproachPlate::SectionEnum)BinaryData->CNSection;
	CNPage=BinaryData->CNPage;
	BVSection=(ApproachPlate::SectionEnum)BinaryData->BVSection;
	BVPage=BinaryData->BVPage;
	ProcUID=BinaryData->ProcUID;
	TwoColored=BinaryData->TwoColored;
	Civil=(ApproachPlate::CivilEnum)BinaryData->Civil;

	/* Reset the size to the actual number of characters */
	StateID=StateID.c_str();
	StateName=StateName.c_str();
	CityName=CityName.c_str();
	Name=Name.c_str();
	ImgName=ImgName.c_str();
}
