/*
 * Vifir - An electronic flight bag wannabe utility for aircraft pilots
 * Copyright (C) 2009 David Ludlow
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * dave@adsllc.com, PO Box 13070, Toledo, OH 43623
 */

#ifdef HAVE_CONFIG_H
 #include "../config.h"
#endif

#include <math.h>
#ifndef WIN32
#include <pthread.h>
#include <poll.h>
#endif
#include <iostream>
#include <cerrno>
#include "GPS.h"
#include "Util.h"

GPS::GPS(Config *Cfg)
{
#ifdef HAVE_LIBGPS
	gpsddata=NULL;
#endif
	ActiveFlag=false;
	this->Cfg=Cfg;

	Lon=Lat=Alt=Trk=Spd=NAN;
	LonLatTime=AltTime=VelTime=0;
}

GPS::~GPS()
{
}

bool GPS::Activate() {
#ifdef HAVE_LIBGPS
	switch (Cfg->GetGPSSource())
	{
		case Config::gpsd:
			/* gpsd.mainframe.cx:2947 and gpsd.rellim.com:2947 have gpsd data sources we can connect to for testing */
#if GPSD_API_MAJOR_VERSION >= 5
			if (!gps_open (Cfg->GetgpsdHost().c_str(),Cfg->GetgpsdPort().c_str(), &gpsdata_buf)) {
				gpsddata = &gpsdata_buf;
			} else {
				gpsddata = NULL;
			}
#else
			gpsddata=gps_open(Cfg->GetgpsdHost().c_str(),Cfg->GetgpsdPort().c_str());
#endif
			if (!gpsddata) {
				fprintf(stderr,"Failure in gps_open(%s,%s)\n",Cfg->GetgpsdHost().c_str(),Cfg->GetgpsdPort().c_str());
				return false;
			}
#if GPSD_API_MAJOR_VERSION==3
			gps_query(gpsddata,"w+x\n");
#elif GPSD_API_MAJOR_VERSION>=4
			gps_stream(gpsddata, WATCH_ENABLE, NULL);
#else
# error Unsupported GPSd library version
#endif

			break;
		case Config::Simulated:
			Lon=Cfg->GetGPSSimLon();
			Lat=Cfg->GetGPSSimLat();
			Spd=KT2MPERS(Cfg->GetGPSSimSpd());
			Trk=Cfg->GetGPSSimTrk();
			Alt=FT2METERS(Cfg->GetGPSSimAlt());
			LonLatTime=time(NULL);
			VelTime=LonLatTime;
			AltTime=LonLatTime;
			break;
		case Config::Serial:
			break;
	}

	ActiveFlag=true;
#endif
	return true;
}

void GPS::Deactivate() {
#ifdef HAVE_LIBGPS
	if (gpsddata) {
		gps_close(gpsddata);
#if GPSD_API_MAJOR_VERSION >= 5
		delete gpsddata;
#else
		gpsddata=NULL;
#endif
	}
	ActiveFlag=false;
#endif
}

bool GPS::GetFix(double *Lon, double *Lat, double *Alt, double *Trk, double *Spd) {
	if (Cfg->GetGPSSource()==Config::Simulated)
	{
		time_t deltat;
		deltat=time(NULL)-LonLatTime;

		this->Spd=KT2MPERS(Cfg->GetGPSSimSpd());
		this->Trk=Cfg->GetGPSSimTrk();
		this->Alt=FT2METERS(Cfg->GetGPSSimAlt());

		/* This is a simple linear approximation.  A proper job would handle the
		 * coordinate system projection.  For that matter, it would do something
		 * with great circles too.  Since this is just a simulation, we don't
		 * bother.
		 */
		this->Lon+=(sin(this->Trk*(M_PI/180.0))*this->Spd*(deltat/3600.0))/60.0;
		this->Lat+=(cos(this->Trk*(M_PI/180.0))*this->Spd*(deltat/3600.0))/60.0;

		LonLatTime=time(NULL);
		VelTime=LonLatTime;
		AltTime=LonLatTime;
#ifdef HAVE_LIBGPS
	} else {
		/* Ask gpsd for data until its buffer is empty */
		struct pollfd pfd;
		memset(&pfd,0,sizeof(pfd));
		pfd.fd=gpsddata->gps_fd;
		pfd.events=POLLIN;
		while (poll(&pfd,1,0)>0) {
			if ((pfd.revents|POLLIN)==POLLIN) {
				int result;
				errno=0;
#if GPSD_API_MAJOR_VERSION >= 5
				result=gps_read(gpsddata);
				if (result > 0)
					result=0;
#else
				result=gps_poll(gpsddata);
#endif
				if (result!=0) {
					if (errno==0) {
						/* Connection to gpsd closed */
						fprintf(stderr,"Connection to gpsd lost.\n");
					} else {
						/* Read error */
#if GPSD_API_MAJOR_VERSION >= 5
						perror("gps_read()");
#else
						perror("gps_poll()");
#endif
					}
					Deactivate();
					return false;
				}
			}
			memset(&pfd,0,sizeof(pfd));
			pfd.fd=gpsddata->gps_fd;
			pfd.events=POLLIN;

			if ((gpsddata->fix.mode==MODE_2D || gpsddata->fix.mode==MODE_3D) && !UTIL_isnan(gpsddata->fix.longitude) && !UTIL_isnan(gpsddata->fix.latitude)) {
				LonLatTime=round(gpsddata->fix.time);
				this->Lon=gpsddata->fix.longitude;
				this->Lat=gpsddata->fix.latitude;
			}
			if (gpsddata->fix.mode==MODE_3D && !UTIL_isnan(gpsddata->fix.altitude)) {
				AltTime=round(gpsddata->fix.time);
				this->Alt=gpsddata->fix.altitude;
			}
			if ((gpsddata->fix.mode==MODE_2D || gpsddata->fix.mode==MODE_3D) && !UTIL_isnan(gpsddata->fix.speed) && !UTIL_isnan(gpsddata->fix.track)) {
				VelTime=round(gpsddata->fix.time);
				this->Spd=gpsddata->fix.speed;
				this->Trk=gpsddata->fix.track;
			}
		}
#endif /* HAVE_LIBGPS */
	}

	if (Lon)
		*Lon=this->Lon;
	if (Lat)
		*Lat=this->Lat;
	if (Alt)
		*Alt=this->Alt;
	if (Trk)
		*Trk=this->Trk;
	if (Spd)
		*Spd=this->Spd;

	if (Lon && UTIL_isnan(*Lon))
		return false;
	if (Lat && UTIL_isnan(*Lat))
		return false;
	if (Alt && UTIL_isnan(*Alt))
		return false;
	if (Trk && UTIL_isnan(*Trk))
		return false;
	if (Spd && UTIL_isnan(*Spd))
		return false;

	return true;
}

bool GPS::GetActive() {
	return ActiveFlag;
}
