/*
 * Vifir - An electronic flight bag wannabe utility for aircraft pilots
 * Copyright (C) 2009 David Ludlow
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * dave@adsllc.com, PO Box 13070, Toledo, OH 43623
 */

#include <errno.h>
#include <fcntl.h>
#include <time.h>
#include <sys/time.h>
#include <cstring>
#include <cstdlib>
#include <sstream>
#include <fstream>
#include <iostream>
#include <iomanip>
#include <cassert>
#include <libglademm.h>
#include <glibmm/i18n.h>
#include <gtkmm/messagedialog.h>
#include <gtkmm/main.h>
#include <gtkmm/filechooserdialog.h>
#include <gtkmm/stock.h>
#include "ImportXPlane900Data.h"
#include "Exception.h"
#include "Util.h"
#include "DataStore.h"
#include "Runway.h"
#include "Airport.h"
#include "Frequency.h"
#include "NavAid.h"
#include "EnrouteWaypoint.h"

ImportXPlane900Data::ImportXPlane900Data(Config *Cfg, std::string *GladeFilename, DataStore *Data)
{
	this->Cfg=Cfg;
	this->GladeFilename=GladeFilename;
	this->Data=Data;

	try
	{
		std::string gladefile;

		refXml = Gnome::Glade::Xml::create(*GladeFilename, "ProgressWindow");
	}
	catch (const Gnome::Glade::XmlError& ex)
	{
		Gtk::MessageDialog *message = new Gtk::MessageDialog(_("Error reading data file: ")+*GladeFilename+"\n"+_("Program cannot continue."),
			false,Gtk::MESSAGE_ERROR,Gtk::BUTTONS_CANCEL,true);
		message->run();
		Gtk::Main::quit();
		return;
	}

	refXml->get_widget("ProgressWindow",ProgressWindow);
	refXml->get_widget("SubProgressBar",SubProgressBar);
	refXml->get_widget("ProgressBar",ProgressBar);
	refXml->get_widget("CancelButton",CancelButton);

	ProgressWindow->set_deletable(false);
	CancelButton->signal_clicked().connect(sigc::mem_fun(this,&ImportXPlane900Data::on_CancelButton_clicked));

	CancelFlag=false;
	BytesRead=0;
	BytesAvail=0;
	DataBufLen=0;
	DataBufOfs=0;
	gettimeofday(&LastProgressBarUpdate,NULL);
}

ImportXPlane900Data::~ImportXPlane900Data() {
	delete ProgressWindow;
}

int ImportXPlane900Data::Run()
{
	std::vector<std::string>::iterator Linesi;

	Gtk::FileChooserDialog FileDialog("Browse to the X-Plane 9.00 data", Gtk::FILE_CHOOSER_ACTION_OPEN);
	FileDialog.set_current_folder(Cfg->GetXPlane900DataImportDir());

	/* Add filters */
	Gtk::FileFilter FileFilter;
	FileFilter.set_name("X-Plane 9.00 Data (AptNav??????XP900.zip)");
	FileFilter.add_pattern("AptNav??????XP900.zip");
	FileDialog.add_filter(FileFilter);

	/* Add response buttons */
	FileDialog.add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
	FileDialog.add_button(Gtk::Stock::OPEN, Gtk::RESPONSE_OK);

	if (FileDialog.run()!=Gtk::RESPONSE_OK)
		return false;

	Cfg->SetXPlane900DataImportDir(FileDialog.get_current_folder());

	std::string VolName=FileDialog.get_filename();
	VolName=VolName.substr(VolName.find_last_of("/\\")+1);	/* Strip off the path */
	VolName=VolName.substr(8,2)+VolName.substr(10,2);	/* AptNav200912XP900.zip -> 0912 */
	time_t EffFrom=Util::CalcEffectiveFrom(&VolName);
	time_t EffTo=Util::CalcEffectiveTo(&VolName);
	VolData=Data->GetVolume(&VolName,&EffFrom,&EffTo);
	if (!VolData) {
		throw Exception("Error creating store for volume data",Exception::LEVEL_ERROR);
	}

	if (!ConfirmData(&VolName,EffFrom,EffTo))
		return false;

#ifdef HAVE_LIBZIP

	/* Process the data file here */
	struct zip *zp=NULL;
	int zerror=0;

	zp=zip_open(FileDialog.get_filename().c_str(),0,&zerror);
	if (!zp) {
		char strbuf[1024];
		zip_error_to_str(strbuf, sizeof(strbuf), zerror, errno);
		Gtk::MessageDialog *message = new Gtk::MessageDialog(
			Glib::ustring("Error opening X-Plane 9.00 data file: ")+strbuf,false,Gtk::MESSAGE_ERROR,
			Gtk::BUTTONS_CANCEL,true);
		message->run();
		delete message;
		return false;
	}

	ImportAirportData(zp,&VolName);
	ImportNavData(zp,&VolName);
	ImportFixData(zp,&VolName);

	zip_close(zp);
	zp=NULL;

	if (VolData) {
		/* The data file doesn't specify a location for the airport itself.
		 * Compute it as the center of the runway thresholds.
		 */
		for (std::vector<Airport*>::iterator Ai=VolData->GetAirports()->begin();Ai!=VolData->GetAirports()->end();Ai++) {
			(*Ai)->ComputeLocByRunways();
		}

		/* Store the imported data */
		std::string Filename;
		Filename=Cfg->GetDataStoreDir()+"/datastore.vifir";
		Util::ShellExpand(&Filename);
		try {
			VolData->Save();
		}
		catch (Exception& Ex) {
			Ex.Show();
			return false;
		}
	}
#endif /* HAVE_LIBZIP */

	return true;
}

bool ImportXPlane900Data::GetDataLine(std::string *Line) {
	/* Find the next newline */
	ssize_t i;
	for (i=DataBufOfs;i<DataBufLen;i++) {
		if (i==DataBufOfs && (DataBuf[i]=='\r' || DataBuf[i]=='\n')) {
			/* Ignore empty lines */
			DataBufOfs++;
			if (DataBufOfs>=DataBufLen)
				return false;
		} else {
			if (DataBuf[i]=='\n' || DataBuf[i]=='\r') {
				/* We found our newline */
				DataBuf[i]=0;
				*Line=(char*)&(DataBuf[DataBufOfs]);
				DataBufOfs=i+1;
				return true;
			}
		}
	}
	return false;
}

bool ImportXPlane900Data::CheckMetaData(std::string *Line, int FormatCode, std::string *Volume) {
	/* Example input: 850 Version - data cycle 2009.12, build ... */

	char *tok;
	char *buf;
	std::string Year, Index;

	/* Copy to a normal C buffer for strtok_r() to work with */
	buf=(char*)malloc(Line->length()+1);
	if (!buf)
		return false;
	strncpy(buf,Line->c_str(),Line->length());
	buf[Line->length()]=0;

	if (!(tok=strtok(buf," \t")))
		return false;
	if (FormatCode!=atoi(tok))
		return false;
	if (!(tok=strtok(NULL," \t")))
		return false;
	if (!(tok=strtok(NULL," \t")))
		return false;
	if (!(tok=strtok(NULL," \t")))
		return false;
	if (!(tok=strtok(NULL," \t")))
		return false;
	if (!(tok=strtok(NULL,".")))
		return false;
	Year=&(tok[2]);	/* Skip the first two digits of */
	if (!(tok=strtok(NULL,",")))
		return false;
	Index=tok;

	if (Volume->compare(Year+Index)!=0)
		return false;

	return true;
}

enum ImportXPlane900Data::RowCodeEnum ImportXPlane900Data::GetRowCode(std::string *Line) {
	return (enum RowCodeEnum)atoi(Line->c_str());
}

void ImportXPlane900Data::on_CancelButton_clicked() {
	CancelFlag=true;
}

/* Allow the main loop to run (and the GUI to update from time to time */
void ImportXPlane900Data::ScheduleMainLoop() {
	struct timeval CurrTime;
	double Last,Curr;

	Last=LastProgressBarUpdate.tv_sec+LastProgressBarUpdate.tv_usec/1000000.0;
	gettimeofday(&CurrTime,NULL);
	Curr=CurrTime.tv_sec+CurrTime.tv_usec/1000000.0;
	if (Curr-Last>0.01) {
		Gtk::Main::iteration(false);
		gettimeofday(&LastProgressBarUpdate,NULL);
	}
}

bool ImportXPlane900Data::ConfirmData(std::string *VolName,time_t EffFrom,time_t EffTo) {
	struct tm bdtime;
	std::ostringstream EffFromOStr,EffToOStr;
	bool result=false;

	try {
		Util::GmTime_r(&bdtime,EffFrom);
	}
	catch (Exception& e) {
		e.Show();
		return false;
	}

	EffFromOStr<<"Eff From="<<std::setfill('0')<<
		std::setw(2)<<bdtime.tm_hour<<std::setw(2)<<bdtime.tm_min<<"Z "<<
		std::setw(2)<<(bdtime.tm_mon+1)<<"/"<<
		std::setw(2)<<bdtime.tm_mday<<"/"<<
		std::setw(2)<<(bdtime.tm_year-100)<<"\n";

	try {
		Util::GmTime_r(&bdtime,EffTo);
	}
	catch (Exception& e) {
		e.Show();
		return false;
	}

	EffToOStr<<"Eff To="<<std::setfill('0')<<
		std::setw(2)<<bdtime.tm_hour<<std::setw(2)<<bdtime.tm_min<<"Z "<<
		std::setw(2)<<(bdtime.tm_mon+1)<<"/"<<
		std::setw(2)<<bdtime.tm_mday<<"/"<<
		std::setw(2)<<(bdtime.tm_year-100)<<"\n";

	Gtk::MessageDialog *message = new Gtk::MessageDialog(
		"Volume "+(*VolName)+"\n"+
		EffFromOStr.str()+EffToOStr.str(),false,Gtk::MESSAGE_QUESTION,
		Gtk::BUTTONS_OK_CANCEL,true);
	message->set_title("Confirm");

	result=(message->run()==Gtk::RESPONSE_OK);
	delete message;

	return result;
}

bool ImportXPlane900Data::ImportAirportData(struct zip *zp, std::string *Volume) {
#ifdef HAVE_LIBZIP
	int zerror=0;
	struct zip_stat zstat;

	zerror=zip_stat(zp, "Resources/default scenery/default apt dat/Earth nav data/apt.dat", 0, &zstat);
	if (zerror<0) {
		Gtk::MessageDialog *message = new Gtk::MessageDialog(
			Glib::ustring("Error reading apt.dat info: ")+zip_strerror(zp),false,Gtk::MESSAGE_ERROR,
			Gtk::BUTTONS_CANCEL,true);
		message->run();
		delete message;
		zip_close(zp);
		return false;
	}

	ssize_t DataFileLen=zstat.size;
	ssize_t DataFileProcessed=0;

	struct zip_file *zfp=NULL;
	zfp=zip_fopen(zp,"Resources/default scenery/default apt dat/Earth nav data/apt.dat",0);
	if (!zfp) {
		Gtk::MessageDialog *message = new Gtk::MessageDialog(
			Glib::ustring("Error opening apt.dat file: ")+zip_strerror(zp),false,Gtk::MESSAGE_ERROR,
			Gtk::BUTTONS_CANCEL,true);
		message->run();
		delete message;
		zip_close(zp);
		return false;
	}

	SubProgressBar->set_text("Airport data");
	ProgressBar->set_text("Total");
	ProgressWindow->show();

	int bytes=1;
	ssize_t LineCount=0;
	DataBufLen=0;
	DataBufOfs=0;
	Airport *AP=NULL;
	Runway *RY=NULL, *OtherEnd=NULL;
	Frequency *FQ=NULL;
	while (bytes>0 && !CancelFlag) {
		bytes=zip_fread(zfp, &(DataBuf[DataBufLen]), sizeof(DataBuf)-DataBufLen);
		if (bytes>0) {
			/* Process the data */
			DataBufLen+=bytes;

			std::string Line;
			while (GetDataLine(&Line) && !CancelFlag) {
				if (LineCount==0) {
					/* Ignore. */
				} else if (LineCount==1) {
					if (!CheckMetaData(&Line,850,Volume)) {
						Gtk::MessageDialog *message = new Gtk::MessageDialog(
							Glib::ustring("Invalid apt.dat file."),false,Gtk::MESSAGE_ERROR,
							Gtk::BUTTONS_CANCEL,true);
						message->run();
						delete message;
						CancelFlag=true;
					}
				} else {
					switch (GetRowCode(&Line)) {
						case ROWCODE_AIRPORT:
							AP=new Airport();
							if (!AP->ParseXPlane850(&Line)) {
								delete AP;
								AP=NULL;
							} else {
								/* Store the Airport in the volume data */
								VolData->GetAirports()->push_back(AP);
							}
							break;
						case ROWCODE_RUNWAY:
							assert(AP!=NULL);	/* The last processed airport should still be around */
							if (AP) {
								RY=new Runway();
								OtherEnd=new Runway();
								if (!RY->ParseXPlane850(&Line,OtherEnd,AP->GetICAO(),AP->GetElevation())) {
									delete RY;
									RY=NULL;
									delete OtherEnd;
									OtherEnd=NULL;
								} else {
									/* Update the airport */
									AP->AddRunway(RY);
									AP->AddRunway(OtherEnd);
								}
							}
							break;
						case ROWCODE_FREQ_WX:
						case ROWCODE_FREQ_UNICOM:
						case ROWCODE_FREQ_CLD:
						case ROWCODE_FREQ_GROUND:
						case ROWCODE_FREQ_TOWER:
						case ROWCODE_FREQ_APP:
						case ROWCODE_FREQ_DEP:
							assert(AP!=NULL);	/* The last processed airport should still be around */
							if (AP) {
								FQ=new Frequency();
								if (!FQ->ParseXPlane850(&Line,AP->GetICAO())) {
									delete FQ;
									FQ=NULL;
								} else {
									/* Update the airport */
									AP->AddFrequency(FQ);
								}
							}
							break;
						default:
							/* Ignore */
							break;
					}
				}
				LineCount++;
			}

			/* There's not enough data in the buffer to complete another line.
			 * Prepare the buffer to accept more data.
			 */
			if (DataBufLen>DataBufOfs) {
				memmove(DataBuf,&(DataBuf[DataBufOfs]),DataBufLen-DataBufOfs);
			}
			DataBufLen-=DataBufOfs;
			DataFileProcessed+=DataBufOfs;
			DataBufOfs=0;

			SubProgressBar->set_fraction((double)DataFileProcessed/DataFileLen);
			ProgressBar->set_fraction(ParseAirportsProgressFrac*DataFileProcessed/DataFileLen);
			ScheduleMainLoop();
		}
	}

	if (bytes==-1) {
		Gtk::MessageDialog *message = new Gtk::MessageDialog(
			Glib::ustring("Error reading apt.dat file: ")+zip_file_strerror(zfp),false,Gtk::MESSAGE_ERROR,
			Gtk::BUTTONS_CANCEL,true);
		message->run();
		zip_fclose(zfp);
		zip_close(zp);
		delete message;
	}

	zip_fclose(zfp);

#endif	/* HAVE_LIBZIP */
	return true;
}

bool ImportXPlane900Data::ImportNavData(struct zip *zp, std::string *Volume) {
#ifdef HAVE_LIBZIP
	int zerror=0;
	struct zip_stat zstat;

	zerror=zip_stat(zp, "Resources/default data/earth_nav.dat", 0, &zstat);
	if (zerror<0) {
		Gtk::MessageDialog *message = new Gtk::MessageDialog(
			Glib::ustring("Error reading earth_nav.dat info: ")+zip_strerror(zp),false,Gtk::MESSAGE_ERROR,
			Gtk::BUTTONS_CANCEL,true);
		message->run();
		delete message;
		zip_close(zp);
		return false;
	}

	ssize_t DataFileLen=zstat.size;
	ssize_t DataFileProcessed=0;

	struct zip_file *zfp=NULL;
	zfp=zip_fopen(zp,"Resources/default data/earth_nav.dat",0);
	if (!zfp) {
		Gtk::MessageDialog *message = new Gtk::MessageDialog(
			Glib::ustring("Error opening earth_nav.dat file: ")+zip_strerror(zp),false,Gtk::MESSAGE_ERROR,
			Gtk::BUTTONS_CANCEL,true);
		message->run();
		delete message;
		zip_close(zp);
		return false;
	}

	SubProgressBar->set_text("Nav data");
	ProgressBar->set_text("Total");
	ProgressWindow->show();

	int bytes=1;
	ssize_t LineCount=0;
	DataBufLen=0;
	DataBufOfs=0;
	NDB *NB=NULL;
	VOR *VR=NULL;
	ILS *IS=NULL;
	Localizer *LC=NULL;
	while (bytes>0 && !CancelFlag) {
		bytes=zip_fread(zfp, &(DataBuf[DataBufLen]), sizeof(DataBuf)-DataBufLen);
		if (bytes>0) {
			/* Process the data */
			DataBufLen+=bytes;

			std::string Line;
			while (GetDataLine(&Line) && !CancelFlag) {
				if (LineCount==0) {
					/* Ignore. */
				} else if (LineCount==1) {
					if (!CheckMetaData(&Line,810,Volume)) {
						Gtk::MessageDialog *message = new Gtk::MessageDialog(
							Glib::ustring("Invalid earth_nav.dat file."),false,Gtk::MESSAGE_ERROR,
							Gtk::BUTTONS_CANCEL,true);
						message->run();
						delete message;
						CancelFlag=true;
					}
				} else {
					switch (GetRowCode(&Line)) {
						case ROWCODE_NDB:
							NB=new NDB();
							if (!NB->ParseXPlane810(&Line)) {
								delete NB;
								NB=NULL;
							} else {
								/* Store the NDB in the volume data */
								VolData->GetNavAids()->push_back(NB);
							}
							break;
						case ROWCODE_VOR:
							VR=new VOR();
							if (!VR->ParseXPlane810(&Line)) {
								delete VR;
								VR=NULL;
							} else {
								/* Store the VOR in the volume data */
								VolData->GetNavAids()->push_back(VR);
							}
							break;
						case ROWCODE_LOCALIZER_ILS:
							IS=new ILS();
							if (!IS->ParseXPlane810(&Line)) {
								delete IS;
								IS=NULL;
							} else {
								/* Store the ILS in the volume data */
								VolData->GetNavAids()->push_back(IS);
							}
							break;
						case ROWCODE_LOCALIZER_ONLY:
							LC=new Localizer();
							if (!VR->ParseXPlane810(&Line)) {
								delete LC;
								LC=NULL;
							} else {
								/* Store the Localizer in the volume data */
								VolData->GetNavAids()->push_back(LC);
							}
							break;
						/* Ignore glideslopes, marker beacons, and standalone DME's */
						default:
							/* Ignore */
							break;
					}
				}
				LineCount++;
			}

			/* There's not enough data in the buffer to complete another line.
			 * Prepare the buffer to accept more data.
			 */
			if (DataBufLen>DataBufOfs) {
				memmove(DataBuf,&(DataBuf[DataBufOfs]),DataBufLen-DataBufOfs);
			}
			DataBufLen-=DataBufOfs;
			DataFileProcessed+=DataBufOfs;
			DataBufOfs=0;

			SubProgressBar->set_fraction((double)DataFileProcessed/DataFileLen);
			ProgressBar->set_fraction(ParseAirportsProgressFrac+ParseNavProgressFrac*DataFileProcessed/DataFileLen);
			ScheduleMainLoop();
		}
	}

	if (bytes==-1) {
		Gtk::MessageDialog *message = new Gtk::MessageDialog(
			Glib::ustring("Error reading earth_nav.dat file: ")+zip_file_strerror(zfp),false,Gtk::MESSAGE_ERROR,
			Gtk::BUTTONS_CANCEL,true);
		message->run();
		zip_fclose(zfp);
		zip_close(zp);
		delete message;
	}

	zip_fclose(zfp);

#endif	/* HAVE_LIBZIP */
	return true;
}

bool ImportXPlane900Data::ImportFixData(struct zip *zp, std::string *Volume) {
#ifdef HAVE_LIBZIP
	int zerror=0;
	struct zip_stat zstat;

	zerror=zip_stat(zp, "Resources/default data/earth_fix.dat", 0, &zstat);
	if (zerror<0) {
		Gtk::MessageDialog *message = new Gtk::MessageDialog(
			Glib::ustring("Error reading earth_fix.dat info: ")+zip_strerror(zp),false,Gtk::MESSAGE_ERROR,
			Gtk::BUTTONS_CANCEL,true);
		message->run();
		delete message;
		zip_close(zp);
		return false;
	}

	ssize_t DataFileLen=zstat.size;
	ssize_t DataFileProcessed=0;

	struct zip_file *zfp=NULL;
	zfp=zip_fopen(zp,"Resources/default data/earth_fix.dat",0);
	if (!zfp) {
		Gtk::MessageDialog *message = new Gtk::MessageDialog(
			Glib::ustring("Error opening earth_fix.dat file: ")+zip_strerror(zp),false,Gtk::MESSAGE_ERROR,
			Gtk::BUTTONS_CANCEL,true);
		message->run();
		delete message;
		zip_close(zp);
		return false;
	}

	SubProgressBar->set_text("Fix data");
	ProgressBar->set_text("Total");
	ProgressWindow->show();

	int bytes=1;
	ssize_t LineCount=0;
	DataBufLen=0;
	DataBufOfs=0;

	while (bytes>0 && !CancelFlag) {
		bytes=zip_fread(zfp, &(DataBuf[DataBufLen]), sizeof(DataBuf)-DataBufLen);
		if (bytes>0) {
			/* Process the data */
			DataBufLen+=bytes;

			std::string Line;
			while (GetDataLine(&Line) && !CancelFlag) {
				if (LineCount==0) {
					/* Ignore. */
				} else if (LineCount==1) {
					if (!CheckMetaData(&Line,600,Volume)) {
						Gtk::MessageDialog *message = new Gtk::MessageDialog(
							Glib::ustring("Invalid earth_fix.dat file."),false,Gtk::MESSAGE_ERROR,
							Gtk::BUTTONS_CANCEL,true);
						message->run();
						delete message;
						CancelFlag=true;
					}
				} else {
					EnrouteWaypoint *EW=new EnrouteWaypoint();
					if (!EW->ParseXPlane600(&Line)) {
						delete EW;
						EW=NULL;
					} else {
						/* Store the Enroute Waypoint in the volume data */
						VolData->GetEnrouteWaypoints()->push_back(EW);
					}
				}
				LineCount++;
			}

			/* There's not enough data in the buffer to complete another line.
			 * Prepare the buffer to accept more data.
			 */
			if (DataBufLen>DataBufOfs) {
				memmove(DataBuf,&(DataBuf[DataBufOfs]),DataBufLen-DataBufOfs);
			}
			DataBufLen-=DataBufOfs;
			DataFileProcessed+=DataBufOfs;
			DataBufOfs=0;

			SubProgressBar->set_fraction((double)DataFileProcessed/DataFileLen);
			ProgressBar->set_fraction(ParseAirportsProgressFrac+ParseNavProgressFrac+
					ParseFixProgressFrac*DataFileProcessed/DataFileLen);
			ScheduleMainLoop();
		}
	}

	if (bytes==-1) {
		Gtk::MessageDialog *message = new Gtk::MessageDialog(
			Glib::ustring("Error reading earth_fix.dat file: ")+zip_file_strerror(zfp),false,Gtk::MESSAGE_ERROR,
			Gtk::BUTTONS_CANCEL,true);
		message->run();
		zip_fclose(zfp);
		zip_close(zp);
		delete message;
	}

	zip_fclose(zfp);

#endif	/* HAVE_LIBZIP */
	return true;
}
